"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.helpers
from cfnlint.data import AdditionalSpecs
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class OnlyOne(CloudFormationLintRule):
    """Check Properties Resource Configuration"""

    id = "E2523"
    shortdesc = "Check Properties that need only one of a list of properties"
    description = (
        "Making sure CloudFormation properties "
        + "that require only one property from a list. "
        + "One has to be specified."
    )
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["resources"]

    def __init__(self):
        """Init"""
        super().__init__()
        onlyonespec = cfnlint.helpers.load_resource(AdditionalSpecs, "OnlyOne.json")
        self.resource_types_specs = onlyonespec["ResourceTypes"]
        self.property_types_specs = onlyonespec["PropertyTypes"]
        for resource_type_spec in self.resource_types_specs:
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in self.property_types_specs:
            self.resource_sub_property_types.append(property_type_spec)

    def check(self, properties, onlyoneprops, path, cfn):
        """Check itself"""
        matches = []

        for onlyoneprop in onlyoneprops:
            for safe_properties, safe_path in properties.items_safe(path):
                property_sets = cfn.get_object_without_conditions(
                    safe_properties, onlyoneprop
                )
                for property_set in property_sets:
                    count = 0
                    for prop in onlyoneprop:
                        if prop in property_set["Object"]:
                            count += 1

                    if count != 1:
                        if property_set["Scenario"] is None:
                            message = "Only one of [{0}] should be specified for {1}"
                            matches.append(
                                RuleMatch(
                                    path,
                                    message.format(
                                        ", ".join(map(str, onlyoneprop)),
                                        "/".join(map(str, safe_path)),
                                    ),
                                )
                            )
                        else:
                            scenario_text = " and ".join(
                                [
                                    f'when condition "{k}" is {v}'
                                    for (k, v) in property_set["Scenario"].items()
                                ]
                            )
                            message = "Only one of [{0}] should be specified {1} at {2}"
                            matches.append(
                                RuleMatch(
                                    path,
                                    message.format(
                                        ", ".join(map(str, onlyoneprop)),
                                        scenario_text,
                                        "/".join(map(str, safe_path)),
                                    ),
                                )
                            )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        onlyoneprops = self.property_types_specs.get(property_type, {})
        matches.extend(self.check(properties, onlyoneprops, path, cfn))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        onlyoneprops = self.resource_types_specs.get(resource_type, {})
        matches.extend(self.check(properties, onlyoneprops, path, cfn))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class FindInMap(CloudFormationLintRule):
    """Check if FindInMap values are correct"""

    id = "E1011"
    shortdesc = "FindInMap validation of configuration"
    description = "Making sure the function is a list of appropriate config"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-findinmap.html"
    tags = ["functions", "findinmap"]

    supported_functions = ["Fn::FindInMap", "Ref"]
    enhanced_supported_functions = [
        "Fn::FindInMap",
        "Ref",
        "Fn::Select",
        "Fn::Split",
        "Fn::Join",
        "Fn::Sub",
        "Fn::If",
        "Fn::Length",
        "Fn::ToJsonString",
        "Fn::Sub",
    ]
    default_value_key = "DefaultValue"

    def check_dict(self, obj, tree):
        """
        Check that obj is a dict with Ref as the only key
        Mappings only support Ref inside them
        """
        matches = []

        if isinstance(obj, dict):
            if len(obj) == 1:
                for key_name, _ in obj.items():
                    if key_name not in self.supported_functions:
                        message = "FindInMap only supports [{0}] functions at {1}"
                        matches.append(
                            RuleMatch(
                                tree[:] + [key_name],
                                message.format(
                                    ", ".join(map(str, self.supported_functions)),
                                    "/".join(map(str, tree)),
                                ),
                            )
                        )
            else:
                message = (
                    "FindInMap only supports an object of 1 of [{0}] functions at {1}"
                )
                matches.append(
                    RuleMatch(
                        tree[:],
                        message.format(
                            ", ".join(map(str, self.supported_functions)),
                            "/".join(map(str, tree)),
                        ),
                    )
                )

        return matches

    def map_name(self, map_name, mappings, tree):
        """Check the map name"""
        matches = []
        if isinstance(map_name, (str, dict)):
            if isinstance(map_name, dict):
                matches.extend(self.check_dict(map_name, tree[:] + [0]))
            else:
                if map_name not in mappings:
                    message = "Map Name {0} does not exist for {0}"
                    matches.append(
                        RuleMatch(
                            tree[:] + [0],
                            message.format(map_name, "/".join(map(str, tree))),
                        )
                    )
        else:
            message = "Map Name should be a {0}, or string at {1}"
            matches.append(
                RuleMatch(
                    tree[:] + [0],
                    message.format(
                        ", ".join(map(str, self.supported_functions)),
                        "/".join(map(str, tree)),
                    ),
                )
            )

        return matches

    def match_key(self, key, tree, key_name, key_index):
        """Check the validity of a key"""
        matches = []
        if isinstance(key, (str, int)):
            return matches
        if isinstance(key, dict):
            matches.extend(self.check_dict(key, tree[:] + [key_index]))
        else:
            message = "FindInMap {0} should be a {1}, string, or int at {2}"
            matches.append(
                RuleMatch(
                    tree[:] + [key_index],
                    message.format(
                        key_name,
                        ", ".join(map(str, self.supported_functions)),
                        "/".join(map(str, tree)),
                    ),
                )
            )

        return matches

    def validate_intrinsic_function(self, obj, tree):
        matches = []
        if not isinstance(obj, dict):
            return matches
        error_message = (
            "FindInMap only supports an object of 1 of [{0}] functions at {1}"
        )
        if len(obj) != 1:
            matches.append(
                RuleMatch(
                    tree[:],
                    error_message.format(
                        "/".join(map(str, self.enhanced_supported_functions)),
                        "/".join(map(str, tree)),
                    ),
                )
            )
            return matches
        function_name = next(iter(obj))
        if function_name not in self.enhanced_supported_functions:
            matches.append(
                RuleMatch(
                    tree[:],
                    error_message.format(
                        "/".join(map(str, self.enhanced_supported_functions)),
                        "/".join(map(str, tree)),
                    ),
                )
            )
        return matches

    def validate_enhanced_map_name(self, map_name, mappings, tree):
        matches = []
        if isinstance(map_name, str):
            if map_name not in mappings:
                error_message = "Map Name {0} does not exist for {1}"
                matches.append(
                    RuleMatch(
                        tree[:] + [0],
                        error_message.format(map_name, "/".join(map(str, tree))),
                    )
                )
        elif isinstance(map_name, dict):
            return self.validate_intrinsic_function(map_name, tree)
        else:
            error_message = "Map Name should be an intrinsic function that resolves to a string, or a string at {0}"
            matches.append(
                RuleMatch(tree[:] + [0], error_message.format("/".join(map(str, tree))))
            )
        return matches

    def validate_enhanced_key(self, key, tree, key_name, key_index):
        matches = []
        if isinstance(key, (str, int)):
            return matches
        if isinstance(key, dict):
            return self.validate_intrinsic_function(key, tree)
        error_message = "FindInMap {0} should be a string, int or an intrinsic function resolves to a string or int at {1}"
        matches.append(
            RuleMatch(
                tree[:] + [key_index],
                error_message.format(key_name, "/".join(map(str, tree))),
            )
        )
        return matches

    def validate_default_value_shape(self, default_value, tree):
        matches = []
        error_message = "Fn::FindInMap default value must be an object whose key is 'DefaultValue' at {0}"
        if (not isinstance(default_value, dict)) or (len(default_value) != 1):
            matches.append(
                RuleMatch(tree[:], error_message.format("/".join(map(str, tree))))
            )
            return matches
        key = next(iter(default_value))
        if key != self.default_value_key:
            matches.append(
                RuleMatch(tree[:], error_message.format("/".join(map(str, tree))))
            )
        return matches

    def validate_enhanced_default_value(self, default_value_dict, tree):
        matches = []
        matches.extend(self.validate_default_value_shape(default_value_dict, tree))
        if not isinstance(default_value_dict, dict):
            return matches
        if self.default_value_key not in default_value_dict:
            return matches
        default_value = default_value_dict[self.default_value_key]
        if isinstance(default_value, (int, float, str)):
            return matches
        if isinstance(default_value, dict):
            matches.extend(self.validate_intrinsic_function(default_value, tree))
        if isinstance(default_value, list):
            for value in default_value:
                if isinstance(value, dict):
                    matches.extend(self.validate_intrinsic_function(value, tree))
                elif isinstance(value, (int, float, str)):
                    continue
                else:
                    error_message = "Elements in Fn::FindInMap default value must be a string, number or an intrinsic function resolves to string at {0}"
                    matches.append(
                        RuleMatch(
                            tree[:], error_message.format("/".join(map(str, tree)))
                        )
                    )
        return matches

    def validate_enhanced_find_in_map(self, find_in_map, cfn):
        mappings = cfn.get_mappings()
        tree = find_in_map[:-1]
        map_obj = find_in_map[-1]
        matches = []
        if not isinstance(map_obj, list):
            error_message = "FindInMap is a list with 3 or 4 values for {0}"
            matches.append(
                RuleMatch(tree[:], error_message.format("/".join(map(str, tree))))
            )
            return matches
        if len(map_obj) in (3, 4):
            map_name = map_obj[0]
            first_key = map_obj[1]
            second_key = map_obj[2]
            matches.extend(self.validate_enhanced_map_name(map_name, mappings, tree))
            matches.extend(self.validate_enhanced_key(first_key, tree, "first key", 1))
            matches.extend(
                self.validate_enhanced_key(second_key, tree, "second key", 2)
            )
            if len(map_obj) == 4:
                default_value = map_obj[3]
                matches.extend(
                    self.validate_enhanced_default_value(default_value, tree)
                )
        else:
            error_message = "FindInMap is a list with 3 or 4 values for {0}"
            matches.append(
                RuleMatch(tree[:] + [1], error_message.format("/".join(map(str, tree))))
            )
        return matches

    def _match_with_language_extensions(self, cfn):
        matches = []
        find_in_maps = cfn.transform_pre["Fn::FindInMap"]
        for in_map in find_in_maps:
            matches.extend(
                self.validate_enhanced_find_in_map(find_in_map=in_map, cfn=cfn)
            )
        return matches

    def _match(self, cfn):
        matches = []

        findinmaps = cfn.search_deep_keys("Fn::FindInMap")
        mappings = cfn.get_mappings()
        for findinmap in findinmaps:
            tree = findinmap[:-1]
            map_obj = findinmap[-1]
            if not isinstance(map_obj, list):
                message = "FindInMap is a list with 3 values for {0}"
                matches.append(RuleMatch(tree[:], message.format("/".join(tree))))
                continue
            if len(map_obj) == 3:
                map_name = map_obj[0]
                first_key = map_obj[1]
                second_key = map_obj[2]

                matches.extend(self.map_name(map_name, mappings, tree))
                matches.extend(self.match_key(first_key, tree, "first key", 1))
                matches.extend(self.match_key(second_key, tree, "second key", 2))

            else:
                message = "FindInMap is a list with 3 values for {0}"
                matches.append(
                    RuleMatch(tree[:] + [1], message.format("/".join(map(str, tree))))
                )

        return matches

    def match(self, cfn):
        if cfn.has_language_extensions_transform():
            return self._match_with_language_extensions(cfn=cfn)
        return self._match(cfn=cfn)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class If(CloudFormationLintRule):
    """Check if Condition exists"""

    id = "E1028"
    shortdesc = "Check Fn::If structure for validity"
    description = "Check Fn::If to make sure its valid.  Condition has to be a string."
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-if"
    tags = ["functions", "if"]

    def match(self, cfn):
        matches = []

        # Build the list of functions
        iftrees = cfn.search_deep_keys("Fn::If")

        # Get the conditions used in the functions
        for iftree in iftrees:
            ifs = iftree[-1]
            if isinstance(ifs, list):
                if_condition = ifs[0]
                if len(ifs) != 3:
                    message = "Fn::If must be a list of 3 elements."
                    matches.append(RuleMatch(iftree[:-1], message))
                if not isinstance(if_condition, str):
                    message = "Fn::If first element must be a condition and a string."
                    matches.append(RuleMatch(iftree[:-1] + [0], message))
            else:
                message = "Fn::If must be a list of 3 elements."
                matches.append(RuleMatch(iftree[:-1], message))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import datetime

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Base(CloudFormationLintRule):
    """Check Base Template Settings"""

    id = "E1001"
    shortdesc = "Basic CloudFormation Template Configuration"
    description = "Making sure the basic CloudFormation template components are properly configured"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["base"]

    required_keys = ["Resources"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.config_definition = {"sections": {"default": "", "type": "string"}}
        self.configure()

    def _validate_version(self, template):
        results = []
        valid_version = "2010-09-09"
        if "AWSTemplateFormatVersion" in template:
            version = template.get("AWSTemplateFormatVersion")
            if not isinstance(version, (str, datetime.date)):
                message = "AWSTemplateFormatVersion only valid value is {0}"
                results.append(
                    RuleMatch(
                        ["AWSTemplateFormatVersion"], message.format(valid_version)
                    )
                )
            else:
                if (
                    version != valid_version
                    and version
                    != datetime.datetime.strptime(valid_version, "%Y-%m-%d").date()
                ):
                    message = "AWSTemplateFormatVersion only valid value is {0}"
                    results.append(
                        RuleMatch(
                            ["AWSTemplateFormatVersion"], message.format(valid_version)
                        )
                    )
        return results

    def _validate_transform(self, transforms):
        results = []
        if not isinstance(transforms, (list, str)):
            message = "Transform has to be a list or string"
            results.append(RuleMatch(["Transform"], message.format()))
            return results

        if not isinstance(transforms, list):
            transforms = [transforms]

        return results

    def match(self, cfn):
        matches = []

        top_level = []
        for x in cfn.template:
            top_level.append(x)
            if x not in cfn.sections and x != self.config["sections"]:
                message = "Top level template section {0} is not valid"
                matches.append(RuleMatch([x], message.format(x)))

        for y in self.required_keys:
            if y not in top_level:
                message = "Missing top level template section {0}"
                matches.append(RuleMatch([y], message.format(y)))

        matches.extend(self._validate_version(cfn.template))
        if "Transform" in cfn.template:
            matches.extend(self._validate_transform(cfn.template.get("Transform")))
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RuleTargetsLimit(CloudFormationLintRule):
    """Check State Machine Definition"""

    id = "E3021"
    shortdesc = "Check Events Rule Targets are less than or equal to 5"
    description = "CloudWatch Events Rule can only support up to 5 targets"
    source_url = "https://docs.aws.amazon.com/AmazonCloudWatch/latest/events/cloudwatch_limits_cwe.html"
    tags = ["resources", "events"]
    max_count = 5

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types = ["AWS::Events::Rule"]
        self.limits = {}

    def initialize(self, cfn):
        self.limits = {}

    # pylint: disable=W0613
    def check_value(self, value, path):
        """Count them up"""

        resource_name = path[1]
        if len(path) > 4:
            if path[4] == "Fn::If":
                resource_name = f"{path[1]}.{path[5]}"

        if resource_name not in self.limits:
            self.limits[resource_name] = {"count": 0, "path": path[:-1]}

        self.limits[resource_name]["count"] += 1
        return []

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []
        matches.extend(
            cfn.check_value(
                obj=properties,
                key="Targets",
                path=path[:],
                check_value=self.check_value,
            )
        )

        for _, limit in self.limits.items():
            if limit["count"] > self.max_count:
                message = "An Events Rule can have up to {0} Targets"
                matches.append(RuleMatch(limit["path"], message.format(self.max_count)))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.common import alphanumeric_name


class Name(CloudFormationLintRule):
    """Check if Outputs are named correctly"""

    id = "E6004"
    shortdesc = "Outputs have appropriate names"
    description = "Check if Outputs are properly named (A-Za-z0-9)"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs"]

    def match(self, cfn):
        return alphanumeric_name(cfn, "Outputs")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.helpers
from cfnlint.helpers import REGISTRY_SCHEMAS
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class Configuration(CloudFormationLintRule):
    """Check Base Resource Configuration"""

    id = "E3001"
    shortdesc = "Basic CloudFormation Resource Check"
    description = (
        "Making sure the basic CloudFormation resources are properly configured"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["resources"]

    def _check_resource(self, cfn: Template, resource_name, resource_values):
        """Check Resource"""

        valid_attributes = [
            "Condition",
            "CreationPolicy",
            "DeletionPolicy",
            "DependsOn",
            "Metadata",
            "Properties",
            "Type",
            "UpdatePolicy",
            "UpdateReplacePolicy",
        ]

        valid_custom_attributes = [
            "Condition",
            "DeletionPolicy",
            "DependsOn",
            "Metadata",
            "Properties",
            "Type",
            "UpdateReplacePolicy",
            "Version",
        ]

        matches = []
        if not isinstance(resource_values, dict):
            message = "Resource not properly configured at {0}"
            matches.append(
                RuleMatch(["Resources", resource_name], message.format(resource_name))
            )
            return matches

        # validate condition is a string
        condition = resource_values.get("Condition", "")
        if not isinstance(condition, str):
            message = "Condition for resource {0} should be a string"
            matches.append(
                RuleMatch(
                    ["Resources", resource_name, "Condition"],
                    message.format(resource_name),
                )
            )

        resource_type = resource_values.get("Type", "")
        if not isinstance(resource_type, str):
            message = "Type has to be a string at {0}"
            matches.append(
                RuleMatch(
                    ["Resources", resource_name],
                    message.format("/".join(["Resources", resource_name])),
                )
            )
            return matches

        # Type is valid continue analysis
        if (
            resource_type.startswith("Custom::")
            or resource_type == "AWS::CloudFormation::CustomResource"
        ):
            check_attributes = valid_custom_attributes
        else:
            check_attributes = valid_attributes

        for property_key, _ in resource_values.items():
            if property_key not in check_attributes:
                message = "Invalid resource attribute {0} for resource {1}"
                matches.append(
                    RuleMatch(
                        ["Resources", resource_name, property_key],
                        message.format(property_key, resource_name),
                    )
                )

        if not resource_type:
            message = "Type not defined for resource {0}"
            matches.append(
                RuleMatch(["Resources", resource_name], message.format(resource_name))
            )
        elif not isinstance(resource_type, str):
            message = "Type has to be a string at {0}"
            matches.append(
                RuleMatch(
                    ["Resources", resource_name],
                    message.format("/".join(["Resources", resource_name])),
                )
            )
        else:
            self.logger.debug("Check resource types by region...")
            for region, specs in cfnlint.helpers.RESOURCE_SPECS.items():
                if region in cfn.regions:
                    if condition:
                        if False in cfn.conditions.build_scenerios_on_region(
                            condition, region
                        ):
                            continue
                    if resource_type not in specs[
                        "ResourceTypes"
                    ] and resource_type not in [
                        s["typeName"] for s in REGISTRY_SCHEMAS
                    ]:
                        if not resource_type.startswith(
                            ("Custom::", "AWS::Serverless::")
                        ) and not resource_type.endswith("::MODULE"):
                            message = "Invalid or unsupported Type {0} for resource {1} in {2}"
                            matches.append(
                                RuleMatch(
                                    ["Resources", resource_name, "Type"],
                                    message.format(
                                        resource_type, resource_name, region
                                    ),
                                )
                            )

        if "Properties" not in resource_values:
            resource_spec = cfnlint.helpers.RESOURCE_SPECS[cfn.regions[0]]
            if resource_type in resource_spec["ResourceTypes"]:
                properties_spec = resource_spec["ResourceTypes"][resource_type][
                    "Properties"
                ]
                # pylint: disable=len-as-condition
                if len(properties_spec) > 0:
                    required = 0
                    for _, property_spec in properties_spec.items():
                        if property_spec.get("Required", False):
                            required += 1
                    if required > 0:
                        if (
                            resource_type == "AWS::CloudFormation::WaitCondition"
                            and "CreationPolicy" in resource_values.keys()
                        ):
                            self.logger.debug(
                                "Exception to required properties section as CreationPolicy is defined."
                            )
                        else:
                            message = "Properties not defined for resource {0}"
                            matches.append(
                                RuleMatch(
                                    ["Resources", resource_name],
                                    message.format(resource_name),
                                )
                            )

        return matches

    def match(self, cfn):
        matches = []

        resources = cfn.template.get("Resources", {})
        if not isinstance(resources, dict):
            message = "Resource not properly configured"
            matches.append(RuleMatch(["Resources"], message))
        else:
            for resource_name, resource_values in cfn.template.get(
                "Resources", {}
            ).items():
                self.logger.debug(
                    "Validating resource %s base configuration", resource_name
                )
                matches.extend(
                    self._check_resource(cfn, resource_name, resource_values)
                )

        return matches



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""


class Match:
    """Match Classes"""

    def __init__(
        self,
        linenumber,
        columnnumber,
        linenumberend,
        columnnumberend,
        filename,
        rule,
        message=None,
        rulematch_obj=None,
    ):
        """Init"""
        self.linenumber = linenumber
        """Starting line number of the region this match spans"""
        self.columnnumber = columnnumber
        """Starting line number of the region this match spans"""
        self.linenumberend = linenumberend
        """Ending line number of the region this match spans"""
        self.columnnumberend = columnnumberend
        """Ending column number of the region this match spans"""
        self.filename = filename
        """Name of the filename associated with this match, or None if there is no such file"""
        self.rule = rule
        """The rule of this match"""
        self.message = message  # or rule.shortdesc
        """The message of this match"""
        if rulematch_obj:
            for k, v in vars(rulematch_obj).items():
                if not hasattr(self, k):
                    setattr(self, k, v)

    def __repr__(self):
        """Represent"""
        file_str = self.filename + ":" if self.filename else ""
        return f"[{self.rule}] ({self.message}) matched {file_str}{self.linenumber}"

    def __eq__(self, item):
        """Override equal to compare matches"""
        return (self.linenumber, self.columnnumber, self.rule.id, self.message) == (
            item.linenumber,
            item.columnnumber,
            item.rule.id,
            item.message,
        )



class CfnLintError(Exception):
    """
    The base exception class for cfn-lint exceptions.
    :ivar msg: The descriptive message associated with the error.
    """

    fmt = "An unspecified error occurred"

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


class DuplicateRuleError(CfnLintError):
    """
    The data associated with a particular path could not be loaded.
    :ivar data_path: The data path that the user attempted to load.
    """

    fmt = "Rule already included: {rule_id}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint.helpers import REGEX_ALPHANUMERIC, REGEX_IPV4, REGEX_IPV6
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RecordSet(CloudFormationLintRule):
    """Check Route53 Recordset Configuration"""

    id = "E3020"
    shortdesc = "Validate Route53 RecordSets"
    description = "Check if all RecordSets are correctly configured"
    source_url = "https://docs.aws.amazon.com/Route53/latest/DeveloperGuide/ResourceRecordTypes.html"
    tags = ["resources", "route53", "record_set"]

    # Regex generated from https://docs.aws.amazon.com/Route53/latest/DeveloperGuide/DomainNameFormat.html
    REGEX_DOMAINNAME = re.compile(
        r"^[a-zA-Z0-9\!\"\#\$\%\&\'\(\)\*\+\,-\/\:\;\<\=\>\?\@\[\\\]\^\_\`\{\|\}\~\.]+$"
    )
    REGEX_TXT = re.compile(r'^("[^"]{1,255}" *)*"[^"]{1,255}"$')
    REGEX_CNAME_VALIDATIONS = re.compile(r"^.*\.acm-validations\.aws\.?$")

    def count_c_names(self, records, path, cfn):
        """Count C Names"""
        matches = []

        scenarios = cfn.get_object_without_nested_conditions(records, path)
        for scenario in scenarios:
            if len(scenario.get("Object")) > 1:
                scenario = scenario.get("Scenario")
                message = "A CNAME recordset can only contain 1 value"
                if scenario is None:
                    message = "A CNAME recordset can only contain 1 value"
                    matches.append(
                        RuleMatch(path, message.format("/".join(map(str, message))))
                    )
                else:
                    message = "A CNAME recordset can only contain 1 value {0} at {1}"
                    scenario_text = " and ".join(
                        [f'when condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    matches.append(
                        RuleMatch(
                            path,
                            message.format(scenario_text, "/".join(map(str, path))),
                        )
                    )

        return matches

    def check_record(self, value, path, record_type, regex, regex_name):
        matches = []
        if isinstance(value, str):
            if not re.match(regex, value):
                message = record_type + " record ({}) is not a valid " + regex_name
                matches.append(RuleMatch(path, message.format(value)))
        return matches

    def check_a_record(self, value, path):
        return self.check_record(value, path, "A", REGEX_IPV4, "IPv4 address")

    def check_aaaa_record(self, value, path):
        return self.check_record(value, path, "AAAA", REGEX_IPV6, "IPv6 address")

    def check_caa_record(self, value, path):
        """Check CAA record Configuration"""
        matches = []

        if isinstance(value, str):
            # Split the record up to the mandatory settings (flags tag "value")
            items = value.split(" ", 2)
            # Check if the 3 settings are given.
            if len(items) != 3:
                message = 'CAA record must contain 3 settings (flags tag "value"), record contains {} settings.'
                matches.append(RuleMatch(path, message.format(len(items))))
            else:
                # Check the flag value
                if not items[0].isdigit():
                    message = "CAA record flag setting ({}) should be of type Integer."
                    extra_args = {
                        "actual_type": type(items[0]).__name__,
                        "expected_type": int.__name__,
                    }
                    matches.append(
                        RuleMatch(path, message.format(items[0]), **extra_args)
                    )
                else:
                    if int(items[0]) not in [0, 128]:
                        message = "Invalid CAA record flag setting ({}) given, must be 0 or 128."
                        matches.append(RuleMatch(path, message.format(items[0])))

                # Check the tag value
                if not re.match(REGEX_ALPHANUMERIC, items[1]):
                    message = "Invalid CAA record tag setting {}. Value has to be alphanumeric."
                    matches.append(RuleMatch(path, message.format(items[1])))

                # Check the value
                if not items[2].startswith('"') or not items[2].endswith('"'):
                    message = 'CAA record value setting has to be enclosed in double quotation marks (").'
                    matches.append(RuleMatch(path, message))

        return matches

    def check_cname_record(self, value, path):
        """Check CNAME record Configuration"""
        matches = []

        if not isinstance(value, dict):
            if not re.match(self.REGEX_DOMAINNAME, value) and not re.match(
                self.REGEX_CNAME_VALIDATIONS, value
            ):
                # ACM Route 53 validation uses invalid CNAMEs starting with `_`,
                # special-case them rather than complicate the regex.
                message = "CNAME record ({}) does not contain a valid domain name"
                matches.append(RuleMatch(path, message.format(value)))

        return matches

    def check_mx_record(self, value, path):
        """Check MX record Configuration"""
        matches = []

        if isinstance(value, str):
            # Split the record up to the mandatory settings (priority domainname)
            items = value.split(" ")

            # Check if the 3 settings are given.
            if len(items) != 2:
                message = "MX record must contain 2 settings (priority domainname), record contains {} settings."
                matches.append(RuleMatch(path, message.format(len(items), value)))
            else:
                # Check the priority value
                if not items[0].isdigit():
                    message = (
                        "MX record priority setting ({}) should be of type Integer."
                    )
                    extra_args = {
                        "actual_type": type(items[0]).__name__,
                        "expected_type": int.__name__,
                    }
                    matches.append(
                        RuleMatch(path, message.format(items[0], value), **extra_args)
                    )
                else:
                    if not 0 <= int(items[0]) <= 65535:
                        message = "Invalid MX record priority setting ({}) given, must be between 0 and 65535."
                        matches.append(RuleMatch(path, message.format(items[0], value)))

                # Check the domainname value
                if not re.match(self.REGEX_DOMAINNAME, items[1]):
                    matches.append(RuleMatch(path, message.format(items[1])))

        return matches

    def check_ns_record(self, value, path):
        return self.check_record(
            value, path, "NS", self.REGEX_DOMAINNAME, "domain name"
        )

    def check_ptr_record(self, value, path):
        return self.check_record(
            value, path, "PTR", self.REGEX_DOMAINNAME, "domain name"
        )

    def check_txt_record(self, value, path):
        """Check TXT record Configuration"""
        matches = []

        if not isinstance(value, dict) and not re.match(self.REGEX_TXT, value):
            message = (
                "TXT record is not structured as one or more items up to 255 characters "
                "enclosed in double quotation marks at {0}"
            )
            matches.append(
                RuleMatch(
                    path,
                    (message.format("/".join(map(str, path)))),
                )
            )

        return matches

    def check_recordset(self, path, recordset, cfn):
        """Check record configuration"""

        matches = []
        recordset_type = recordset.get("Type")

        # Skip Intrinsic functions
        if not isinstance(recordset_type, dict):
            if not recordset.get("AliasTarget"):
                # If no Alias is specified, ResourceRecords has to be specified
                if not recordset.get("ResourceRecords"):
                    return matches
                # Record type specific checks
                if recordset_type == "A":
                    matches.extend(
                        cfn.check_value(
                            recordset,
                            "ResourceRecords",
                            path[:],
                            check_value=self.check_a_record,
                        )
                    )
                elif recordset_type == "AAAA":
                    matches.extend(
                        cfn.check_value(
                            recordset,
                            "ResourceRecords",
                            path[:],
                            check_value=self.check_aaaa_record,
                        )
                    )
                elif recordset_type == "CAA":
                    matches.extend(
                        cfn.check_value(
                            recordset,
                            "ResourceRecords",
                            path[:],
                            check_value=self.check_caa_record,
                        )
                    )
                elif recordset_type == "CNAME":
                    matches.extend(
                        self.count_c_names(
                            recordset.get("ResourceRecords"),
                            path[:] + ["ResourceRecords"],
                            cfn,
                        )
                    )
                    matches.extend(
                        cfn.check_value(
                            recordset,
                            "ResourceRecords",
                            path[:],
                            check_value=self.check_cname_record,
                        )
                    )
                elif recordset_type == "MX":
                    matches.extend(
                        cfn.check_value(
                            recordset,
                            "ResourceRecords",
                            path[:],
                            check_value=self.check_mx_record,
                        )
                    )
                elif recordset_type == "NS":
                    matches.extend(
                        cfn.check_value(
                            recordset,
                            "ResourceRecords",
                            path[:],
                            check_value=self.check_ns_record,
                        )
                    )
                elif recordset_type == "PTR":
                    matches.extend(
                        cfn.check_value(
                            recordset,
                            "ResourceRecords",
                            path[:],
                            check_value=self.check_ptr_record,
                        )
                    )
                elif recordset_type == "TXT":
                    matches.extend(
                        cfn.check_value(
                            recordset,
                            "ResourceRecords",
                            path[:],
                            check_value=self.check_txt_record,
                        )
                    )
            else:
                if recordset.get("TTL"):
                    matches.append(
                        RuleMatch(
                            path + ["TTL"], "TTL is not allowed for Alias records"
                        )
                    )

        return matches

    def match(self, cfn):
        """Check RecordSets and RecordSetGroups Properties"""

        matches = []

        recordsets = cfn.get_resources(["AWS::Route53::RecordSet"])

        for name, recordset in recordsets.items():
            path = ["Resources", name, "Properties"]

            if isinstance(recordset, dict):
                props = recordset.get("Properties")
                if props:
                    matches.extend(self.check_recordset(path, props, cfn))

        recordsetgroups = cfn.get_resource_properties(
            ["AWS::Route53::RecordSetGroup", "RecordSets"]
        )

        for recordsetgroup in recordsetgroups:
            path = recordsetgroup["Path"]
            value = recordsetgroup["Value"]
            if isinstance(value, list):
                for index, recordset in enumerate(value):
                    tree = path[:] + [index]
                    matches.extend(self.check_recordset(tree, recordset, cfn))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class GetAz(CloudFormationLintRule):
    """Check if GetAz values are correct"""

    id = "E1015"
    shortdesc = "GetAz validation of parameters"
    description = "Making sure the GetAz function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-getavailabilityzones.html"
    tags = ["functions", "getaz"]

    def match(self, cfn):
        matches = []

        getaz_objs = cfn.search_deep_keys("Fn::GetAZs")

        for getaz_obj in getaz_objs:
            getaz_value = getaz_obj[-1]
            if isinstance(getaz_value, str):
                if getaz_value != "" and getaz_value not in cfn.regions:
                    message = (
                        "GetAZs should be of empty or string of valid region for {0}"
                    )
                    matches.append(
                        RuleMatch(
                            getaz_obj[:-1],
                            message.format("/".join(map(str, getaz_obj[:-1]))),
                        )
                    )
            elif isinstance(getaz_value, dict):
                if len(getaz_value) == 1:
                    if isinstance(getaz_value, dict):
                        for key, value in getaz_value.items():
                            if key != "Ref" or value != "AWS::Region":
                                message = (
                                    "GetAZs should be of Ref to AWS::Region for {0}"
                                )
                                matches.append(
                                    RuleMatch(
                                        getaz_obj[:-1],
                                        message.format(
                                            "/".join(map(str, getaz_obj[:-1]))
                                        ),
                                    )
                                )
                    else:
                        message = "GetAZs should be of Ref to AWS::Region for {0}"
                        matches.append(
                            RuleMatch(
                                getaz_obj[:-1],
                                message.format("/".join(map(str, getaz_obj[:-1]))),
                            )
                        )
                else:
                    message = "GetAZs should be of Ref to AWS::Region for {0}"
                    matches.append(
                        RuleMatch(
                            getaz_obj[:-1],
                            message.format("/".join(map(str, getaz_obj[:-1]))),
                        )
                    )
        return matches
